@@ -1,6 +1,6 @@
 /*
   Copyright (C) 2012 David Robillard <d@drobilla.net>
-    
+
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
@@ -16,11 +16,14 @@
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
-#include <stddef.h>
-
 #include "autowah.h"
 #include "plugin_lv2.h"
 
+#include "lv2/core/lv2.h"
+
+#include <cstddef>
+#include <cstdint>
+
 extern "C" {
 
 static LV2_Handle
