@@ -5,7 +5,7 @@ import sys
 import time
 
 from waflib import Configure, ConfigSet, Build, Context, Logs, Options, Utils
-from waflib.TaskGen import feature, before, after
+from waflib.TaskGen import feature, before, after, after_method
 
 NONEMPTY = -10
 
@@ -20,11 +20,33 @@ else:
 # import preproc
 # preproc.go_absolute = True
 
+
 @feature('c', 'cxx')
 @after('apply_incpaths')
 def include_config_h(self):
     self.env.append_value('INCPATHS', self.bld.bldnode.abspath())
 
+
+def _set_system_headers(self, varname):
+    if 'AUTOWAF_SYSTEM_PKGS' in self.env and not self.env.MSVC_COMPILER:
+        for lib in self.uselib:
+            if lib in self.env.AUTOWAF_SYSTEM_PKGS:
+                for include in self.env['INCLUDES_' + lib]:
+                    self.env.append_unique(varname, ['-isystem%s' % include])
+
+
+@feature('c')
+@after_method('apply_incpaths')
+def set_system_headers_c(self):
+    _set_system_headers(self, 'CFLAGS')
+
+
+@feature('cxx')
+@after_method('apply_incpaths')
+def set_system_headers_cxx(self):
+    _set_system_headers(self, 'CXXFLAGS')
+
+
 class OptionsContext(Options.OptionsContext):
     def __init__(self, **kwargs):
         super(OptionsContext, self).__init__(**kwargs)
@@ -39,7 +61,8 @@ class OptionsContext(Options.OptionsContext):
             group.add_option('--' + name, action='store_true',
                              dest=name.replace('-', '_'), help=desc)
 
-def set_options(opt, debug_by_default=False):
+
+def set_options(opt):
     "Add standard autowaf options"
     opts = opt.get_option_group('Configuration options')
 
@@ -60,15 +83,11 @@ def set_options(opt, debug_by_default=False):
                     help="HTML documentation [default: DATADIR/doc]")
 
     # Build options
-    if debug_by_default:
-        opts.add_option('--optimize', action='store_false', default=True,
-                        dest='debug', help="build optimized binaries")
-    else:
-        opts.add_option('-d', '--debug', action='store_true', default=False,
-                        dest='debug', help="build debuggable binaries")
-        opts.add_option('--pardebug', action='store_true', default=False,
-                        dest='pardebug',
-                        help="build debug libraries with D suffix")
+    opts.add_option('-d', '--debug', action='store_true', default=False,
+                    dest='debug', help="build debuggable binaries")
+    opts.add_option('--pardebug', action='store_true', default=False,
+                    dest='pardebug',
+                    help="build debug libraries with D suffix")
 
     opts.add_option('-s', '--strict', action='store_true', default=False,
                     dest='strict',
@@ -84,14 +103,11 @@ def set_options(opt, debug_by_default=False):
     # Test options
     if hasattr(Context.g_module, 'test'):
         test_opts = opt.add_option_group('Test options', '')
-        opts.add_option('-T', '--test', action='store_true', dest='build_tests',
-                        help='build unit tests')
+        opts.add_option('-T', '--test', action='store_true',
+                        dest='build_tests', help='build unit tests')
         opts.add_option('--no-coverage', action='store_true',
                         dest='no_coverage',
                         help='do not instrument code for test coverage')
-        test_opts.add_option('--wrapper', type='string',
-                             dest='test_wrapper',
-                             help='command prefix for tests (e.g. valgrind)')
         test_opts.add_option('--test-filter', type='string',
                              dest='test_filter',
                              help='regular expression for tests to run')
@@ -100,6 +116,10 @@ def set_options(opt, debug_by_default=False):
     run_opts = opt.add_option_group('Run options')
     run_opts.add_option('--cmd', type='string', dest='cmd',
                         help='command to run from build directory')
+    run_opts.add_option('--wrapper', type='string',
+                        dest='wrapper',
+                        help='command prefix for running executables')
+
 
 class ConfigureContext(Configure.ConfigurationContext):
     """configures the project"""
@@ -111,7 +131,6 @@ class ConfigureContext(Configure.ConfigurationContext):
 
         super(ConfigureContext, self).__init__(**kwargs)
         self.run_env = ConfigSet.ConfigSet()
-        self.system_include_paths = set()
 
     def pre_recurse(self, node):
         if len(self.stack_path) == 1:
@@ -120,12 +139,6 @@ class ConfigureContext(Configure.ConfigurationContext):
 
     def store(self):
         self.env.AUTOWAF_RUN_ENV = self.run_env.get_merged_dict()
-        for path in sorted(self.system_include_paths):
-            if 'COMPILER_CC' in self.env:
-                self.env.append_value('CFLAGS', ['-isystem', path])
-            if 'COMPILER_CXX' in self.env:
-                self.env.append_value('CXXFLAGS', ['-isystem', path])
-
         super(ConfigureContext, self).store()
 
     def check_pkg(self, *args, **kwargs):
@@ -138,6 +151,7 @@ class ConfigureContext(Configure.ConfigurationContext):
         """Return `path` within the build directory"""
         return str(self.path.get_bld().make_node(path))
 
+
 def get_check_func(conf, lang):
     if lang == 'c':
         return conf.check_cc
@@ -146,6 +160,7 @@ def get_check_func(conf, lang):
     else:
         Logs.error("Unknown header language `%s'" % lang)
 
+
 def check_header(conf, lang, name, define='', mandatory=True):
     "Check for a header"
     check_func = get_check_func(conf, lang)
@@ -156,32 +171,51 @@ def check_header(conf, lang, name, define='', mandatory=True):
     else:
         check_func(header_name=name, mandatory=mandatory)
 
+
 def check_function(conf, lang, name, **args):
     "Check for a function"
     header_names = Utils.to_list(args['header_name'])
     includes = ''.join(['#include <%s>\n' % x for x in header_names])
+    return_type = args['return_type'] if 'return_type' in args else 'int'
+    arg_types = args['arg_types'] if 'arg_types' in args else ''
+
     fragment = '''
 %s
-int main() { return !(void(*)())(%s); }
-''' % (includes, name)
+
+typedef %s (*Func)(%s);
+
+int main(void) {
+    static const Func ptr = %s;
+    (void)ptr;
+    return 0;
+}
+''' % (includes, return_type, arg_types, name)
 
     check_func  = get_check_func(conf, lang)
     args['msg'] = 'Checking for %s' % name
+    if lang + 'flags' not in args:
+        args[lang + 'flags'] = check_flags(conf, conf.env.CFLAGS)
+
     check_func(fragment=fragment, **args)
 
+
 def nameify(name):
     return (name.replace('/', '_').replace('++', 'PP')
             .replace('-', '_').replace('.', '_'))
 
+
 def check_pkg(conf, spec, **kwargs):
     "Check for a package iff it hasn't been checked for yet"
 
-    if (kwargs['uselib_store'].lower() in conf.env['AUTOWAF_LOCAL_LIBS'] or
-        kwargs['uselib_store'].lower() in conf.env['AUTOWAF_LOCAL_HEADERS']):
+    uselib_store = kwargs['uselib_store']
+    is_local = (uselib_store.lower() in conf.env['AUTOWAF_LOCAL_LIBS'] or
+                uselib_store.lower() in conf.env['AUTOWAF_LOCAL_HEADERS'])
+
+    if is_local:
         return
 
     import re
-    match = re.match('([^ ]*) >= [0-9\.]*', spec)
+    match = re.match(r'([^ ]*) >= [0-9\.]*', spec)
     args = []
     if match:
         name = match.group(1)
@@ -192,7 +226,6 @@ def check_pkg(conf, spec, **kwargs):
         Logs.error("Invalid package spec: %s" % spec)
 
     found = None
-    pkg_var_name = 'PKG_' + name.replace('-', '_')
     pkg_name = name
     args += kwargs.get('args', [])
 
@@ -203,7 +236,7 @@ def check_pkg(conf, spec, **kwargs):
         if found:
             pkg_name += 'D'
 
-        args['mandatory'] = mandatory  # Unsmash mandatory arg
+        args['mandatory'] = kwargs['mandatory']  # Unsmash mandatory arg
 
     if not found:
         found = conf.check_cfg(package=spec,
@@ -211,8 +244,8 @@ def check_pkg(conf, spec, **kwargs):
                                **kwargs)
 
     if not conf.env.MSVC_COMPILER and 'system' in kwargs and kwargs['system']:
-        conf.system_include_paths.update(
-            conf.env['INCLUDES_' + nameify(kwargs['uselib_store'])])
+        conf.env.append_unique('AUTOWAF_SYSTEM_PKGS', uselib_store)
+
 
 def normpath(path):
     if sys.platform == 'win32':
@@ -220,6 +253,245 @@ def normpath(path):
     else:
         return os.path.normpath(path)
 
+
+# Almost all GCC warnings common to C and C++
+gcc_common_warnings = [
+  # '-Waggregate-return', # Pretty esoteric, and not in clang
+  '-Waggressive-loop-optimizations',
+  '-Wall',
+  '-Walloc-zero',
+  '-Walloca',
+  # '-Walloca-larger-than=',
+  '-Wattribute-alias',
+  '-Wattributes',
+  '-Wbuiltin-declaration-mismatch',
+  '-Wbuiltin-macro-redefined',
+  '-Wcast-align',
+  '-Wcast-align=strict',
+  '-Wcast-qual',
+  '-Wconversion',
+  '-Wcoverage-mismatch',
+  '-Wcpp',
+  '-Wdate-time',
+  '-Wdeprecated',
+  '-Wdeprecated-declarations',
+  '-Wdisabled-optimization',
+  '-Wdiv-by-zero',
+  '-Wdouble-promotion',
+  '-Wduplicated-branches',
+  '-Wduplicated-cond',
+  '-Wextra',
+  '-Wfloat-equal',
+  '-Wformat-signedness',
+  '-Wnormalized',
+  # '-Wframe-larger-than=',
+  '-Wfree-nonheap-object',
+  '-Whsa',
+  '-Wif-not-aligned',
+  '-Wignored-attributes',
+  '-Winline',
+  '-Wint-to-pointer-cast',
+  '-Winvalid-memory-model',
+  '-Winvalid-pch',
+  # '-Wlarger-than=',
+  '-Wlogical-op',
+  '-Wlto-type-mismatch',
+  '-Wmissing-declarations',
+  '-Wmissing-include-dirs',
+  '-Wmultichar',
+  '-Wnull-dereference',
+  '-Wodr',
+  '-Woverflow',
+  '-Wpacked',
+  '-Wpacked-bitfield-compat',
+  '-Wpadded',
+  '-Wpedantic',
+  '-Wpointer-compare',
+  '-Wpragmas',
+  '-Wredundant-decls',
+  '-Wreturn-local-addr',
+  '-Wscalar-storage-order',
+  '-Wshadow',
+  '-Wshift-count-negative',
+  '-Wshift-count-overflow',
+  '-Wshift-negative-value',
+  '-Wshift-overflow=2',
+  '-Wsizeof-array-argument',
+  '-Wstack-protector',
+  # '-Wstack-usage=',
+  '-Wstrict-aliasing',
+  '-Wstrict-overflow',
+  '-Wsuggest-attribute=cold',
+  '-Wsuggest-attribute=const',
+  '-Wsuggest-attribute=format',
+  '-Wsuggest-attribute=malloc',
+  '-Wsuggest-attribute=noreturn',
+  '-Wsuggest-attribute=pure',
+  '-Wswitch-bool',
+  '-Wnormalized',
+  # '-Wswitch-default', # Redundant with Wswitch and not in clang
+  '-Wswitch-enum',
+  '-Wswitch-unreachable',
+  '-Wsync-nand',
+  # '-Wsystem-headers',
+  '-Wtrampolines',
+  '-Wundef',
+  '-Wunused-macros',
+  '-Wunused-result',
+  '-Wvarargs',
+  '-Wvector-operation-performance',
+  '-Wvla',
+  # '-Wvla-larger-than=',
+  '-Wwrite-strings',
+]
+
+# Almost all C-specific GCC warnings, except those for ancient (pre-C99) C
+gcc_c_warnings = [
+  '-Wbad-function-cast',
+  '-Wc++-compat',
+  # '-Wc90-c99-compat',
+  '-Wc99-c11-compat',
+  # '-Wdeclaration-after-statement',
+  '-Wdesignated-init',
+  '-Wdiscarded-array-qualifiers',
+  '-Wdiscarded-qualifiers',
+  '-Wincompatible-pointer-types',
+  '-Wint-conversion',
+  '-Wjump-misses-init',
+  '-Wmissing-prototypes',
+  '-Wnested-externs',
+  '-Wold-style-definition',
+  '-Woverride-init-side-effects',
+  '-Wpointer-to-int-cast',
+  '-Wstrict-prototypes',
+  # '-Wtraditional',
+  # '-Wtraditional-conversion',
+  # '-Wunsuffixed-float-constants',
+]
+
+# Almost all C++-specific GCC warnings, except those about common feature use
+gcc_cxx_warnings = [
+  '-Wconditionally-supported',
+  '-Wconversion-null',
+  '-Wctor-dtor-privacy',
+  '-Wdelete-incomplete',
+  '-Weffc++',
+  '-Wextra-semi',
+  '-Winherited-variadic-ctor',
+  '-Winvalid-offsetof',
+  '-Wliteral-suffix',
+  '-Wmultiple-inheritance',
+  # '-Wnamespaces',
+  '-Wnoexcept',
+  '-Wnon-template-friend',
+  '-Wnon-virtual-dtor',
+  '-Wold-style-cast',
+  '-Woverloaded-virtual',
+  '-Wplacement-new=2',
+  '-Wpmf-conversions',
+  '-Wregister',
+  '-Wsign-promo',
+  '-Wstrict-null-sentinel',
+  '-Wsubobject-linkage',
+  '-Wsuggest-final-methods',
+  '-Wsuggest-final-types',
+  '-Wsuggest-override',
+  '-Wsynth',
+  # '-Wtemplates',
+  '-Wterminate',
+  '-Wuseless-cast',
+  '-Wvirtual-inheritance',
+  '-Wvirtual-move-assign',
+  '-Wzero-as-null-pointer-constant',
+]
+
+
+def remove_all_warning_flags(env):
+    """Removes all warning flags except Werror or equivalent"""
+    if 'CC' in env:
+        if 'clang' in env.CC_NAME or 'gcc' in env.CC_NAME:
+            env['CFLAGS'] = [f for f in env['CFLAGS']
+                             if not (f.startswith('-W') and f != '-Werror')]
+        elif 'msvc' in env.CC_NAME:
+            env['CFLAGS'] = [f for f in env['CFLAGS']
+                             if not (f.startswith('/W') and f != '/WX')]
+
+    if 'CXX' in env:
+        if 'clang' in env.CXX_NAME or 'gcc' in env.CXX_NAME:
+            env['CXXFLAGS'] = [f for f in env['CXXFLAGS']
+                               if not (f.startswith('-W') and f != '-Werror')]
+        elif 'msvc' in env.CXX_NAME:
+            env['CXXFLAGS'] = [f for f in env['CXXFLAGS']
+                               if not (f.startswith('/W') and f != '/WX')]
+
+
+def enable_all_warnings(env):
+    """Enables all known warnings"""
+    if 'CC' in env:
+        if 'clang' in env.CC_NAME:
+            env.append_unique('CFLAGS', ['-Weverything'])
+        elif 'gcc' in env.CC_NAME:
+            env.append_unique('CFLAGS', gcc_common_warnings)
+            env.append_unique('CFLAGS', gcc_c_warnings)
+        elif env.MSVC_COMPILER:
+            env.append_unique('CFLAGS', ['/Wall'])
+        else:
+            Logs.warn('Unknown compiler "%s", not enabling warnings' % env.CC_NAME)
+
+    if 'CXX' in env:
+        if 'clang' in env.CXX_NAME:
+            env.append_unique('CXXFLAGS', ['-Weverything',
+                                           '-Wno-c++98-compat',
+                                           '-Wno-c++98-compat-pedantic'])
+        elif 'gcc' in env.CXX_NAME:
+            env.append_unique('CXXFLAGS', gcc_common_warnings)
+            env.append_unique('CXXFLAGS', gcc_cxx_warnings)
+        elif env.MSVC_COMPILER:
+            env.append_unique('CXXFLAGS', ['/Wall'])
+        else:
+            Logs.warn('Unknown compiler "%s", not enabling warnings' % env.CXX_NAME)
+
+
+def set_warnings_as_errors(env):
+    if 'CC' in env:
+        if 'clang' in env.CC_NAME or 'gcc' in env.CC_NAME:
+            env.append_unique('CFLAGS', ['-Werror'])
+        elif env.MSVC_COMPILER:
+            env.append_unique('CFLAGS', ['/WX'])
+
+    if 'CXX' in env:
+        if 'clang' in env.CXX_NAME or 'gcc' in env.CXX_NAME:
+            env.append_unique('CXXFLAGS', ['-Werror'])
+        elif env.MSVC_COMPILER:
+            env.append_unique('CXXFLAGS', ['/WX'])
+
+
+def add_compiler_flags(env, lang, compiler_to_flags):
+    """Add compiler-specific flags, for example to suppress warnings.
+
+    The lang argument must be "c", "cxx", or "*" for both.
+
+    The compiler_to_flags argument must be a map from compiler name
+    ("clang", "gcc", or "msvc") to a list of command line flags.
+    """
+
+    if lang == "*":
+        add_compiler_flags(env, 'c', compiler_to_flags)
+        add_compiler_flags(env, 'cxx', compiler_to_flags)
+    else:
+        if lang == 'c':
+            compiler_name = env.CC_NAME
+        elif lang == 'cxx':
+            compiler_name = env.CXX_NAME
+        else:
+            raise Exception('Unknown language "%s"' % lang)
+
+        var_name = lang.upper() + 'FLAGS'
+        for name, flags in compiler_to_flags.items():
+            if name in compiler_name:
+                env.append_value(var_name, flags)
+
+
 def configure(conf):
     def append_cxx_flags(flags):
         conf.env.append_value('CFLAGS', flags)
@@ -260,17 +532,37 @@ def configure(conf):
 
     if Options.options.debug:
         if conf.env['MSVC_COMPILER']:
-            conf.env['CFLAGS']    = ['/Od', '/Z7', '/MTd', '/FS']
-            conf.env['CXXFLAGS']  = ['/Od', '/Z7', '/MTd', '/FS']
+            conf.env['CFLAGS']    = ['/Od', '/Z7']
+            conf.env['CXXFLAGS']  = ['/Od', '/Z7']
             conf.env['LINKFLAGS'] = ['/DEBUG', '/MANIFEST']
         else:
             conf.env['CFLAGS']   = ['-O0', '-g']
             conf.env['CXXFLAGS'] = ['-O0', '-g']
     else:
-        if conf.env['MSVC_COMPILER']:
-            append_cxx_flags(['/MD', '/FS', '/DNDEBUG'])
-        else:
-            append_cxx_flags(['-DNDEBUG'])
+        if 'CFLAGS' not in os.environ:
+            if conf.env['MSVC_COMPILER']:
+                conf.env.append_unique('CFLAGS', ['/O2', '/DNDEBUG'])
+            else:
+                conf.env.append_unique('CFLAGS', ['-O2', '-DNDEBUG'])
+
+        if 'CXXFLAGS' not in os.environ:
+            if conf.env['MSVC_COMPILER']:
+                conf.env.append_unique('CXXFLAGS', ['/O2', '/DNDEBUG'])
+            else:
+                conf.env.append_unique('CXXFLAGS', ['-O2', '-DNDEBUG'])
+
+    if conf.env['MSVC_COMPILER']:
+        conf.env['CFLAGS']   += ['/MD']
+        conf.env['CXXFLAGS'] += ['/MD']
+
+    if Options.options.ultra_strict:
+        Options.options.strict = True
+        remove_all_warning_flags(conf.env)
+        enable_all_warnings(conf.env)
+        if Options.options.werror and 'clang' in conf.env.CC_NAME:
+            conf.env.append_unique('CFLAGS', '-Wno-unknown-warning-option')
+        if Options.options.werror and 'clang' in conf.env.CXX_NAME:
+            conf.env.append_unique('CXXFLAGS', '-Wno-unknown-warning-option')
 
     if conf.env.MSVC_COMPILER:
         Options.options.no_coverage = True
@@ -280,59 +572,31 @@ def configure(conf):
                           '/experimental:external',
                           '/external:W0',
                           '/external:anglebrackets'])
+        conf.env.append_unique('CXXFLAGS', ['/EHsc'])
         conf.env.append_value('LINKFLAGS', '/nologo')
-        if Options.options.strict or Options.options.ultra_strict:
-            ms_strict_flags = ['/Wall',
-                               '/wd4061',
-                               '/wd4200',
-                               '/wd4514',
-                               '/wd4571',
-                               '/wd4625',
-                               '/wd4626',
-                               '/wd4706',
-                               '/wd4710',
-                               '/wd4820',
-                               '/wd5026',
-                               '/wd5027',
-                               '/wd5045']
-            conf.env.append_value('CFLAGS', ms_strict_flags)
-            conf.env.append_value('CXXFLAGS', ms_strict_flags)
-            conf.env.append_value('CXXFLAGS', ['/EHsc'])
-    else:
-        if Options.options.ultra_strict:
-            Options.options.strict = True
-            conf.env.append_value('CFLAGS', ['-Wredundant-decls',
-                                             '-Wstrict-prototypes',
-                                             '-Wmissing-prototypes',
-                                             '-Wcast-qual'])
-            conf.env.append_value('CXXFLAGS', ['-Wcast-qual'])
+    elif Options.options.strict:
+        if conf.env.DEST_OS != "darwin":
+            sanitizing = False
+            for f in conf.env.LINKFLAGS:
+                if f.startswith('-fsanitize'):
+                    sanitizing = True
+                    break;
 
-        if Options.options.strict:
-            conf.env.append_value('CFLAGS', ['-pedantic', '-Wshadow'])
-            if conf.env.DEST_OS != "darwin":
+            if not sanitizing:
                 conf.env.append_value('LINKFLAGS', ['-Wl,--no-undefined'])
-            conf.env.append_value('CXXFLAGS', ['-Wnon-virtual-dtor',
-                                               '-Woverloaded-virtual'])
-            append_cxx_flags(['-Wall',
-                              '-Wcast-align',
-                              '-Wextra',
-                              '-Wmissing-declarations',
-                              '-Wno-unused-parameter',
-                              '-Wno-parentheses',
-                              '-Wstrict-overflow',
-                              '-Wundef',
-                              '-Wwrite-strings',
-                              '-fstrict-overflow'])
 
             # Add less universal flags after checking they work
             extra_flags = ['-Wlogical-op',
                            '-Wsuggest-attribute=noreturn',
                            '-Wunsafe-loop-optimizations']
-            if conf.check_cc(cflags=['-Werror'] + extra_flags, mandatory=False,
+            cflags = flag_check_flags(conf, conf.env.CFLAGS) + extra_flags
+            if conf.check_cc(cflags=cflags,
+                             mandatory=False,
                              msg="Checking for extra C warning flags"):
                 conf.env.append_value('CFLAGS', extra_flags)
             if 'COMPILER_CXX' in conf.env:
-                if conf.check_cxx(cxxflags=['-Werror'] + extra_flags,
+                cxxflags = flag_check_flags(conf, conf.env.CXXFLAGS) + extra_flags
+                if conf.check_cxx(cxxflags=cxxflags,
                                   mandatory=False,
                                   msg="Checking for extra C++ warning flags"):
                     conf.env.append_value('CXXFLAGS', extra_flags)
@@ -351,7 +615,7 @@ def configure(conf):
     try:
         conf.env.BUILD_TESTS = Options.options.build_tests
         conf.env.NO_COVERAGE = Options.options.no_coverage
-        if not Options.options.no_coverage:
+        if conf.env.BUILD_TESTS and not Options.options.no_coverage:
             # Set up unit test code coverage
             if conf.is_defined('CLANG'):
                 for cov in [conf.env.CC[0].replace('clang', 'llvm-cov'),
@@ -359,10 +623,22 @@ def configure(conf):
                     if conf.find_program(cov, var='LLVM_COV', mandatory=False):
                         break
             else:
-                conf.check_cc(lib='gcov', define_name='HAVE_GCOV',
-                              mandatory=False)
-    except Exception:
-        pass  # Test options do not exist
+                if 'CC' in conf.env:
+                    if conf.check_cc(cflags=check_flags(conf, conf.env.CFLAGS),
+                                lib='gcov',
+                                mandatory=False,
+                                uselib_store='GCOV'):
+                        conf.env.HAVE_GCOV = True
+                else:
+                    if conf.check_cxx(cflags=check_flags(conf, conf.env.CXXFLAGS),
+                                 lib='gcov',
+                                 mandatory=False,
+                                 uselib_store='GCOV'):
+                        conf.env.HAVE_GCOV = True
+    except AttributeError:
+        pass # Test options do not exist
+    except Exception as e:
+        Logs.error("error: %s" % e)
 
     # Define version in configuration
     appname = getattr(Context.g_module, Context.APPNAME, 'noname')
@@ -371,8 +647,6 @@ def configure(conf):
     conf.define(defname + '_VERSION', version)
     conf.env[defname + '_VERSION'] = version
 
-    conf.env.prepend_value('CFLAGS', '-I' + os.path.abspath('.'))
-    conf.env.prepend_value('CXXFLAGS', '-I' + os.path.abspath('.'))
 
 def display_summary(conf, msgs=None):
     if len(conf.stack_path) == 1:
@@ -387,16 +661,44 @@ def display_summary(conf, msgs=None):
     if msgs is not None:
         display_msgs(conf, msgs)
 
-def set_c_lang(conf, lang):
+
+def check_flags(conf, flags):
+    if conf.env.MSVC_COMPILER:
+        return []
+
+    # Disable silly attribute warnings that trigger in the generated check code
+    result = []
+    if '-Wsuggest-attribute=const' in flags:
+        result += ['-Wno-suggest-attribute=const']
+    if '-Wsuggest-attribute=pure' in flags:
+        result += ['-Wno-suggest-attribute=pure']
+
+    return result
+
+
+def flag_check_flags(conf, flags):
+    if conf.env.MSVC_COMPILER:
+        return ['/WX'] + check_flags(conf, flags)
+    else:
+        return ['-Werror'] + check_flags(conf, flags)
+
+
+def set_c_lang(conf, lang, **kwargs):
     "Set a specific C language standard, like 'c99' or 'c11'"
     if conf.env.MSVC_COMPILER:
         # MSVC has no hope or desire to compile C99, just compile as C++
         conf.env.append_unique('CFLAGS', ['/TP'])
-    else:
+        return True
+    elif not (lang == 'c99' and '-std=c11' in conf.env.CFLAGS):
         flag = '-std=%s' % lang
-        conf.check(cflags=['-Werror', flag],
-                   msg="Checking for flag '%s'" % flag)
-        conf.env.append_unique('CFLAGS', [flag])
+        if conf.check(features='c cstlib',
+                      cflags=flag_check_flags(conf, conf.env.CFLAGS) + [flag],
+                      msg="Checking for flag '%s'" % flag,
+                      **kwargs):
+            conf.env.append_unique('CFLAGS', [flag])
+            return True
+        return False
+
 
 def set_cxx_lang(conf, lang):
     "Set a specific C++ language standard, like 'c++11', 'c++14', or 'c++17'"
@@ -406,10 +708,11 @@ def set_cxx_lang(conf, lang):
         conf.env.append_unique('CXXFLAGS', ['/std:%s' % lang])
     else:
         flag = '-std=%s' % lang
-        conf.check(cxxflags=['-Werror', flag],
+        conf.check(cxxflags=flag_check_flags(conf, conf.env.CXXFLAGS) + [flag],
                    msg="Checking for flag '%s'" % flag)
         conf.env.append_unique('CXXFLAGS', [flag])
 
+
 def set_modern_c_flags(conf):
     "Use the most modern C language available"
     if 'COMPILER_CC' in conf.env:
@@ -423,6 +726,7 @@ def set_modern_c_flags(conf):
                     conf.env.append_unique('CFLAGS', [flag])
                     break
 
+
 def set_modern_cxx_flags(conf, mandatory=False):
     "Use the most modern C++ language available"
     if 'COMPILER_CXX' in conf.env:
@@ -436,6 +740,7 @@ def set_modern_cxx_flags(conf, mandatory=False):
                     conf.env.append_unique('CXXFLAGS', [flag])
                     break
 
+
 def set_local_lib(conf, name, has_objects):
     var_name = 'HAVE_' + nameify(name.upper())
     conf.define(var_name, 1)
@@ -449,12 +754,14 @@ def set_local_lib(conf, name, has_objects):
             conf.env['AUTOWAF_LOCAL_HEADERS'] = {}
         conf.env['AUTOWAF_LOCAL_HEADERS'][name.lower()] = True
 
+
 def append_property(obj, key, val):
     if hasattr(obj, key):
         setattr(obj, key, getattr(obj, key) + val)
     else:
         setattr(obj, key, val)
 
+
 @feature('c', 'cxx')
 @before('apply_link')
 def version_lib(self):
@@ -465,17 +772,19 @@ def version_lib(self):
         if [x for x in applicable if x in self.features]:
             self.target = self.target + 'D'
 
+
 def set_lib_env(conf,
                 name,
                 version,
                 has_objects=True,
                 include_path=None,
-                lib_path=None):
+                lib_path=None,
+                lib=None):
     "Set up environment for local library as if found via pkg-config."
     NAME         = name.upper()
     major_ver    = version.split('.')[0]
     pkg_var_name = 'PKG_' + name.replace('-', '_') + '_' + major_ver
-    lib_name     = '%s-%s' % (name, major_ver)
+    lib_name     = '%s-%s' % (lib if lib is not None else name, major_ver)
 
     if lib_path is None:
         lib_path = str(conf.path.get_bld())
@@ -495,6 +804,7 @@ def set_lib_env(conf,
     conf.run_env.append_unique(lib_path_name, [lib_path])
     conf.define(NAME + '_VERSION', version)
 
+
 def display_msg(conf, msg, status=None, color=None):
     color = 'CYAN'
     if type(status) == bool and status:
@@ -507,27 +817,38 @@ def display_msg(conf, msg, status=None, color=None):
     Logs.pprint('BOLD', ":", sep='')
     Logs.pprint(color, status)
 
+
 def display_msgs(conf, msgs):
     for k, v in msgs.items():
         display_msg(conf, k, v)
 
+
 def link_flags(env, lib):
     return ' '.join(map(lambda x: env['LIB_ST'] % x,
                         env['LIB_' + lib]))
 
+
 def compile_flags(env, lib):
     return ' '.join(map(lambda x: env['CPPPATH_ST'] % x,
                         env['INCLUDES_' + lib]))
 
+
 def build_pc(bld, name, version, version_suffix, libs, subst_dict={}):
     """Build a pkg-config file for a library.
 
     name           -- uppercase variable name     (e.g. 'SOMENAME')
+                      or path to template without .pc.in extension
     version        -- version string              (e.g. '1.2.3')
     version_suffix -- name version suffix         (e.g. '2')
     libs           -- string/list of dependencies (e.g. 'LIBFOO GLIB')
     """
 
+    if '/' in name:
+        source = '%s.pc.in' % name.lower()
+        name = os.path.basename(name)
+    else:
+        source = '%s.pc.in' % name.lower()
+
     pkg_prefix       = bld.env['PREFIX']
     if len(pkg_prefix) > 1 and pkg_prefix[-1] == '/':
         pkg_prefix = pkg_prefix[:-1]
@@ -550,7 +871,7 @@ def build_pc(bld, name, version, version_suffix, libs, subst_dict={}):
         includedir = includedir.replace(pkg_prefix, '${prefix}')
 
     obj = bld(features='subst',
-              source='%s.pc.in' % name.lower(),
+              source=source,
               target=target,
               install_path=os.path.join(bld.env['LIBDIR'], 'pkgconfig'),
               exec_prefix='${prefix}',
@@ -573,48 +894,15 @@ def build_pc(bld, name, version, version_suffix, libs, subst_dict={}):
 
     obj.__dict__.update(subst_dict)
 
-def make_simple_dox(name):
-    "Clean up messy Doxygen documentation after it is built"
-    name = name.lower()
-    NAME = name.upper()
-    try:
-        top = os.getcwd()
-        os.chdir(build_dir(name, 'doc/html'))
-        page = 'group__%s.html' % name
-        if not os.path.exists(page):
-            return
-        for i in [
-            ['%s_API ' % NAME, ''],
-            ['%s_DEPRECATED ' % NAME, ''],
-            ['group__%s.html' % name, ''],
-            ['&#160;', ''],
-            [r'<script.*><\/script>', ''],
-            [r'<hr\/><a name="details" id="details"><\/a><h2>.*<\/h2>', ''],
-            [r'<link href=\"tabs.css\" rel=\"stylesheet\" type=\"text\/css\"\/>',
-             ''],
-            [r'<img class=\"footer\" src=\"doxygen.png\" alt=\"doxygen\"\/>',
-             'Doxygen']]:
-            os.system("sed -i 's/%s/%s/g' %s" % (i[0], i[1], page))
-        os.rename('group__%s.html' % name, 'index.html')
-        for i in (glob.glob('*.png') +
-                  glob.glob('*.html') +
-                  glob.glob('*.js') +
-                  glob.glob('*.css')):
-            if i != 'index.html' and i != 'style.css':
-                os.remove(i)
-        os.chdir(top)
-        os.chdir(build_dir(name, 'doc/man/man3'))
-        for i in glob.glob('*.3'):
-            os.system("sed -i 's/%s_API //' %s" % (NAME, i))
-        for i in glob.glob('_*'):
-            os.remove(i)
-        os.chdir(top)
-    except Exception as e:
-        Logs.error("Failed to fix up %s documentation: %s" % (name, e))
-    finally:
-        os.chdir(top)
 
-def build_dox(bld, name, version, srcdir, blddir, outdir='', versioned=True):
+def build_dox(bld,
+              name,
+              version,
+              srcdir,
+              blddir,
+              outdir='',
+              versioned=True,
+              install_man=True):
     """Build Doxygen API documentation"""
     if not bld.env['DOCS']:
         return
@@ -648,10 +936,14 @@ def build_dox(bld, name, version, srcdir, blddir, outdir='', versioned=True):
     bld.install_files(
         os.path.join('${DOCDIR}', outname, outdir, 'html'),
         bld.path.get_bld().ant_glob('doc/html/*'))
-    for i in range(1, 8):
-        bld.install_files('${MANDIR}/man%d' % i,
-                          bld.path.get_bld().ant_glob('doc/man/man%d/*' % i,
-                                                      excl='**/_*'))
+
+    if install_man:
+        for i in range(1, 8):
+            bld.install_files(
+                '${MANDIR}/man%d' % i,
+                bld.path.get_bld().ant_glob('doc/man/man%d/*' % i,
+                                            excl='**/_*'))
+
 
 def build_version_files(header_path, source_path, domain, major, minor, micro):
     """Generate version code header"""
@@ -685,6 +977,7 @@ def build_version_files(header_path, source_path, domain, major, minor, micro):
 
     return None
 
+
 def build_i18n_pot(bld, srcdir, dir, name, sources, copyright_holder=None):
     Logs.info('Generating pot file from %s' % name)
     pot_file = '%s.pot' % name
@@ -703,6 +996,7 @@ def build_i18n_pot(bld, srcdir, dir, name, sources, copyright_holder=None):
     Logs.info('Updating ' + pot_file)
     subprocess.call(cmd, cwd=os.path.join(srcdir, dir))
 
+
 def build_i18n_po(bld, srcdir, dir, name, sources, copyright_holder=None):
     pwd = os.getcwd()
     os.chdir(os.path.join(srcdir, dir))
@@ -717,6 +1011,7 @@ def build_i18n_po(bld, srcdir, dir, name, sources, copyright_holder=None):
         subprocess.call(cmd)
     os.chdir(pwd)
 
+
 def build_i18n_mo(bld, srcdir, dir, name, sources, copyright_holder=None):
     pwd = os.getcwd()
     os.chdir(os.path.join(srcdir, dir))
@@ -733,11 +1028,13 @@ def build_i18n_mo(bld, srcdir, dir, name, sources, copyright_holder=None):
         subprocess.call(cmd)
     os.chdir(pwd)
 
+
 def build_i18n(bld, srcdir, dir, name, sources, copyright_holder=None):
     build_i18n_pot(bld, srcdir, dir, name, sources, copyright_holder)
     build_i18n_po(bld, srcdir, dir, name, sources, copyright_holder)
     build_i18n_mo(bld, srcdir, dir, name, sources, copyright_holder)
 
+
 class ExecutionEnvironment:
     """Context that sets system environment variables for program execution"""
     def __init__(self, changes):
@@ -754,7 +1051,7 @@ class ExecutionEnvironment:
         os.environ.update(self.diff)
 
     def __str__(self):
-        return '\n'.join({'%s="%s"' % (k, v) for k, v in self.diff.items()})
+        return '\n'.join(['%s="%s"' % (k, v) for k, v in self.diff.items()])
 
     def __enter__(self):
         return self
@@ -762,6 +1059,7 @@ class ExecutionEnvironment:
     def __exit__(self, type, value, traceback):
         os.environ = self.original_environ
 
+
 class RunContext(Build.BuildContext):
     "runs an executable from the build directory"
     cmd = 'run'
@@ -781,6 +1079,7 @@ class RunContext(Build.BuildContext):
             else:
                 Logs.error("error: Missing --cmd option for run command")
 
+
 def show_diff(from_lines, to_lines, from_filename, to_filename):
     import difflib
     import sys
@@ -795,6 +1094,7 @@ def show_diff(from_lines, to_lines, from_filename, to_filename):
 
     return same
 
+
 def test_file_equals(patha, pathb):
     import filecmp
     import io
@@ -811,12 +1111,14 @@ def test_file_equals(patha, pathb):
         with io.open(pathb, 'rU', encoding='utf-8') as fb:
             return show_diff(fa.readlines(), fb.readlines(), patha, pathb)
 
+
 def bench_time():
-    if hasattr(time, 'perf_counter'): # Added in Python 3.3
+    if hasattr(time, 'perf_counter'):  # Added in Python 3.3
         return time.perf_counter()
     else:
         return time.time()
 
+
 class TestOutput:
     """Test output that is truthy if result is as expected"""
     def __init__(self, expected, result=None):
@@ -829,11 +1131,13 @@ class TestOutput:
 
     __nonzero__ = __bool__
 
+
 def is_string(s):
     if sys.version_info[0] < 3:
         return isinstance(s, basestring)
     return isinstance(s, str)
 
+
 class TestScope:
     """Scope for running tests that maintains pass/fail statistics"""
     def __init__(self, tst, name, defaults):
@@ -884,7 +1188,7 @@ class TestScope:
             with tempfile.TemporaryFile() as stderr:
                 kwargs['stderr'] = stderr
                 output = self.run(test, **kwargs)
-                stderr.seek(0, 2) # Seek to end
+                stderr.seek(0, 2)  # Seek to end
                 return (output if not output else
                         self.run(
                             lambda: stderr.tell() > 0,
@@ -922,9 +1226,9 @@ class TestScope:
         if verbosity > 1:
             self.tst.log_good('RUN     ', name)
 
-        if Options.options.test_wrapper:
+        if Options.options.wrapper:
             import shlex
-            test = shlex.split(Options.options.test_wrapper) + test
+            test = shlex.split(Options.options.wrapper) + test
 
         output = TestOutput(expected)
         with open(os.devnull, 'wb') as null:
@@ -953,6 +1257,7 @@ class TestScope:
 
         return output
 
+
 class TestContext(Build.BuildContext):
     "runs test suite"
     fun = cmd = 'test'
@@ -1001,7 +1306,8 @@ class TestContext(Build.BuildContext):
             Logs.info("Waf: Entering directory `%s'", bld_dir)
             os.chdir(str(bld_dir))
 
-            if not self.env.NO_COVERAGE and str(node.parent) == Context.top_dir:
+            parent_is_top = str(node.parent) == Context.top_dir
+            if not self.env.NO_COVERAGE and parent_is_top:
                 self.clear_coverage()
 
             Logs.info('')
@@ -1088,8 +1394,8 @@ class TestContext(Build.BuildContext):
                 subprocess.call(['lcov', '-z', '-d', str(self.path)],
                                 stdout=log, stderr=log)
 
-        except Exception:
-            Logs.warn('Failed to run lcov to clear old coverage data')
+        except Exception as e:
+            Logs.warn('Failed to run lcov to clear old coverage data (%s)' % e)
 
     def gen_coverage(self):
         """Generate coverage data and report"""
@@ -1119,14 +1425,16 @@ class TestContext(Build.BuildContext):
                 stderr=subprocess.STDOUT).decode('ascii')
 
             import re
-            lines = re.search('lines\.*: (.*)%.*', summary).group(1)
-            functions = re.search('functions\.*: (.*)%.*', summary).group(1)
-            branches = re.search('branches\.*: (.*)%.*', summary).group(1)
-            self.log_good('COVERAGE', '%s%% lines, %s%% functions, %s%% branches',
-                          lines, functions, branches)
+            lines = re.search(r'lines\.*: (.*)%.*', summary).group(1)
+            functions = re.search(r'functions\.*: (.*)%.*', summary).group(1)
+            branches = re.search(r'branches\.*: (.*)%.*', summary).group(1)
+            self.log_good(
+                'COVERAGE', '%s%% lines, %s%% functions, %s%% branches',
+                lines, functions, branches)
+
+        except Exception as e:
+            Logs.warn('Failed to run lcov to generate coverage report (%s)')
 
-        except Exception:
-            Logs.warn('Failed to run lcov to generate coverage report')
 
 class TestGroup:
     def __init__(self, tst, suitename, name, **kwargs):
@@ -1167,6 +1475,7 @@ class TestGroup:
             self.tst.log_bad('-' * 10, '%d/%d tests from %s (%d ms total)',
                              n_passed, scope.n_total, self.label(), duration)
 
+
 def run_ldconfig(ctx):
     should_run = (ctx.cmd == 'install' and
                   not ctx.env['RAN_LDCONFIG'] and
@@ -1182,6 +1491,7 @@ def run_ldconfig(ctx):
         except Exception:
             pass
 
+
 def run_script(cmds):
     for cmd in cmds:
         subprocess.check_call(cmd, shell=True)
